#!/usr/bin/python

'''
DDL for logging tables

  create table ARCH.dbo.SUPER_LOG
  (    [project] varchar(25)
  , [msg] varchar(5000)
  , [success] varchar(1)
  , [date] datetime2
  , [GUID] varchar(50)
  , [ETL_GUID] varchar(50)
  )

'''

from datetime import datetime
from os import sys
from time import strftime
from traceback import format_exc


#This is an important file to have as it handles your constants for all database connections
import constants

import errno
import logging.handlers
import os

import requests
import time
import uuid
from paver.easy import no_help

#This is the library you use to make connections to your server, plus import all the error types
import pymssql
from pymssql import Error, DatabaseError, OperationalError, IntegrityError, DataError, ProgrammingError, \
    NotSupportedError

"""
This section is for communicating with incoming #webhooks (Slack, Teams, etc)
"""


def webhooks(message, channel):
    """
    The function #webhooks(message,channel) sends messages
    to your specified channel. If you have more than one channel, you can use the parameter to specify.

    Args:
      message(str): This is the message to write to the designated channel.
      channel(str): This is the #webhooks channel to write to
    Returns:
        None
    """

    #Load the incoming #webhooks address
    #webhooks_address = constants.#webhooks_address

    if message == '':
        raise ValueError('No message supplied!')
        return

    
    payload = {"text": message, "parse": "full"}
    headers = {"Content-type": "application/json", "Accept": "text/plain"}

    try:
        requests.post(webhooks_address, json=payload, headers=headers)
    except requests.exceptions.RequestException as e:
        print(e)
        supehr_logger.info(e)


def run_scripts(prefix=None, *args):
    ''' Takes a variable number of arguments that represent paths to SQL scripts
        and executes them. Also keeps a log of all actions done with scripts.'''

    # First check to make sure that prefix was defined. If prefix was not defined this indicates a 
    # configuration error in a Python file.
    try:
        if prefix is None:
            raise ValueError('A prefix was not specified for one or more scripts.')

    except ValueError as ve:
        supehr_logger.error('An issue was encountered validating one or more SQL scripts.', exc_info=True)
        arch_cnxn.rollback()
        cursor.close()
        arch_cnxn.close()
        raise
        sys.exit()

    sql_query_array = [""] * len(args)
    i = 0

    for entry in args:
        sql_query_array[i] = constants.sql_file_directory + entry

        i += 1

    def run_etl_decorator(function):
        @no_help  # Suppress this method in paver help menu
        def run_etl(*args, **kwargs):

            ## this is to connect to databases. Only activate one at a time!
            ## If developing from a new machine you may have to install FreeTDS/UnixODBC and adjust parameters in the connect string appropriately.
            arch_cnxn = pymssql.connect(server=server_address, database=db, user=username, password=pw)
            cursor = arch_cnxn.cursor()

            arch_cnxn.commit()

            for query in sql_query_array:
                # Guid for query
                guid = uuid.uuid4().hex

                t_begin = datetime.utcnow()
                print('ETL GUID = ' + ETL_guid)
                print('GUID = ' + guid)
                print(strftime("%Y-%m-%d %H:%M:%S") + ": START " + query)
                supehr_logger.info('ETL GUID=' + ETL_guid)
                supehr_logger.info('GUID=' + guid)
                supehr_logger.info(
                    query + " called from " + function.__name__ + " at " + strftime("%Y-%m-%d %H:%M:%S") + '\n')
                # Log to #webhooks
                #webhooks_message = '\n---------------------\nETL_id: ' + ETL_guid + '\n'
                #webhooks_message += 'Log_id=' + guid + '\n'
                #webhooks_message += prefix + ' ' + query + " \n*STARTED BY* " + function.__name__ + "\n at " + strftime(
                #    "%Y-%m-%d %H:%M:%S") + '\n---------------------\n'
                #webhooks(#webhooks_message,etl_channel)

                # Print the start time that this query was run into a database log file
                cursor.execute("INSERT INTO ARCH.dbo.SUPER_LOG VALUES (%s,%s,%s,%d,%s,%s);", (
                prefix, 'Started ' + query + ' on ' + prefix, 'Y', strftime("%Y-%m-%d %H:%M:%S"), guid, ETL_guid))
                arch_cnxn.commit()

                '''
                # Generate the before counts for i2b2 ETL
                before_sql_cmd = constants.sql_file_directory + 'QA/etl_history_before.sql'
                sql_cmd_before = ""
                for line in open(before_sql_cmd, 'r'):
                    sql_cmd_before += line + " "

                sql_cmd_before = sql_cmd_before.replace('<prefix>', prefix)
                sql_cmd_before = sql_cmd_before.replace('<outer_function_name>', executing_statement)
                sql_cmd_before = sql_cmd_before.replace('<variable>', guid)
                sql_cmd_before = sql_cmd_before.replace('<function_name>', function.__name__)
                sql_cmd_before = sql_cmd_before.replace('<ETL_variable>', ETL_guid)
                sql_cmd_before = sql_cmd_before.replace('<query>', query)

                cursor.execute(sql_cmd_before)
                arch_cnxn.commit()
                
                '''
                sqlcmd = ""
                # The script must be read into memory before it can be sent to the server
                for line in open(query, 'r'):
                    sqlcmd += line + " "

                # String replacement of <prefix> in file with prefix parameter.
                # Allows for generalization of SQL scripts
                sqlcmd = sqlcmd.replace('<prefix>', prefix)

                try:
                    cursor.execute(sqlcmd)
                    arch_cnxn.commit()
                    t_end = datetime.utcnow()
                    print('ETL GUID = ' + ETL_guid)
                    print('GUID = ' + guid)
                    print(strftime("%Y-%m-%d %H:%M:%S") + ": FINISHED " + query)
                    print("Time taken: " + str(t_end - t_begin))
                    supehr_logger.info('ETL GUID=' + ETL_guid)
                    supehr_logger.info('GUID=' + guid)
                    supehr_logger.info(query + " finished at " + strftime("%Y-%m-%d %H:%M:%S") + '\n')
                    supehr_logger.info("Time taken: " + str(t_end - t_begin))
                    supehr_logger.info(str(cursor.rowcount) + " rows affected.")

                    # Print completion of script to log file
                    cursor.execute("INSERT INTO ARCH.dbo.SUPER_LOG VALUES(%s,%s,%s,%d,%s,%s);", (
                    prefix, 'Finished ' + query + ' on ' + prefix, 'Y', strftime("%Y-%m-%d %H:%M:%S"), guid, ETL_guid))
                    arch_cnxn.commit()

                    # Log to #webhooks
                    #webhooks_message = "\n---------------------\nETL_id: " + ETL_guid + '\n'
                    #webhooks_message += "Log_id = " + guid + " \n*FINISHED* at " + strftime("%Y-%m-%d %H:%M:%S") + '\n'
                    #webhooks_message += " Time taken: " + str(t_end - t_begin) + '\n---------------------\n'
                    #webhooks(#webhooks_message, etl_channel)

                    '''
                    # Generate the after counts for i2b2 ETL
                    after_sql_cmd = constants.sql_file_directory + 'QA/etl_history_after.sql'
                    sql_cmd_after = ""
                    for line in open(after_sql_cmd, 'r'):
                        sql_cmd_after += line + " "
                    

                    sql_cmd_after = sql_cmd_after.replace('<prefix>', prefix)
                    sql_cmd_after = sql_cmd_after.replace('<outer_function_name>', executing_statement)
                    sql_cmd_after = sql_cmd_after.replace('<function_name>', function.__name__)
                    sql_cmd_after = sql_cmd_after.replace('<variable>', guid)
                    sql_cmd_after = sql_cmd_after.replace('<ETL_variable>', ETL_guid)
                    sql_cmd_after = sql_cmd_after.replace('<query>', query)

                    cursor.execute(sql_cmd_after)
                    arch_cnxn.commit()

                    '''
                    # Convert days to minutes
                    timeDiff = (t_end - t_begin).seconds

                    if timeDiff >= 300:
                        time.sleep(10)  # Pause for ten seconds to give server time to breathe   

                except (OperationalError, DataError, IntegrityError, ProgrammingError, NotSupportedError, DatabaseError,
                        IndexError, Error) as e:
                    ''' Roll back the database connection, then fail properly '''
                    supehr_logger.info('ETL GUID: ' + ETL_guid)
                    supehr_logger.info('GUID=' + guid)
                    supehr_logger.error("Initiating rollback...")
                    arch_cnxn.rollback()
                    supehr_logger.info('ETL GUID: ' + ETL_guid)
                    supehr_logger.info('GUID=' + guid)
                    supehr_logger.error("An error occurred during the ETL.", exc_info=True)
                    print("Time of error occurrence: " + strftime("%Y-%m-%d %H:%M:%S") + "\n")
                    supehr_logger.error(format_exc())
                    supehr_logger.error("Time of error occurrence: " + strftime("%Y-%m-%d %H:%M:%S") + "\n")
                    cursor.execute("INSERT INTO ARCH.dbo.SUPER_LOG VALUES (%s,%s,%s,%d,%s,%s);",
                                   (prefix, format_exc(), 'N', strftime("%Y-%m-%d %H:%M:%S"), guid, ETL_guid))

                    arch_cnxn.commit()

                    '''
                    # Generate the after counts for i2b2 ETL
                    after_sql_cmd = constants.sql_file_directory + 'QA/etl_history_after.sql'
                    sql_cmd_after = ""
                    for line in open(after_sql_cmd, 'r'):
                        sql_cmd_after += line + " "

                    sql_cmd_after = sql_cmd_after.replace('<prefix>', prefix)
                    sql_cmd_after = sql_cmd_after.replace('<outer_function_name>', executing_statement)
                    sql_cmd_after = sql_cmd_after.replace('<function_name>', function.__name__)
                    sql_cmd_after = sql_cmd_after.replace('<variable>', guid)
                    sql_cmd_after = sql_cmd_after.replace('<ETL_variable>', ETL_guid)
                    sql_cmd_after = sql_cmd_after.replace('<query>', query)

                    cursor.execute(sql_cmd_after)
                    '''
                    cursor.close()
                    arch_cnxn.close()

                    # Log to #webhooks
                    #webhooks_message = "\n---------------------\nETL_id: " + ETL_guid
                    #webhooks_message += "\nLog_id = " + guid + " \nTime of error occurrence: " + strftime("%Y-%m-%d %H:%M:%S")
                    #webhooks_message += format_exc() + "\n---------------------\n"
                    #webhooks(#webhooks_message,etl_channel)

                    # Changes color of output to red so that we can note where errors happen clearer.
                    print('\033[41m')
                    print(format_exc())
                    print('\033[0m')
                    raise
                    # sys.exit()

            cursor.close()
            arch_cnxn.close()

        return run_etl

    return run_etl_decorator


# Make sure a directory exists for storing logs before creating it. Ignore it if it already exists
def make_sure_log_directory_exists():
    '''Check if log directory exists and create it if it doesn't'''
    try:
        os.makedirs('log/')
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def init_logging():
    LOG_FILENAME = './log/test.log'

    ''' 
    
    Set up a specific logger with our desired output level 

    '''
    supehr_logger = logging.getLogger(__name__)
    supehr_logger.setLevel(logging.INFO)

    # Add the log message handler to the logger
    supehr_handler = logging.handlers.RotatingFileHandler(LOG_FILENAME, maxBytes=100000000, backupCount=5)
    supehr_logger.addHandler(supehr_handler)

    supehr_logger.info("\nUser: " + os.getlogin() + "\n")
    supehr_logger.info("\n" + constants.program_name + " Log Output\n---------------------\n")
    supehr_logger.info("ETL_id: " + ETL_guid + ', Executing Statement: paver ' + executing_statement)

    # Log to #webhooks
    #webhooks_message = "\nUser: " + os.getlogin() + "\n"
    #webhooks_message += "\n" + constants.program_name + " " + " Log Output\n---------------------\n"
    #webhooks_message += "ETL_id: " + ETL_guid + ', Executing Statement: paver ' + executing_statement
    #webhooks_message += "Server: " + constants.server_addr
    #webhooks(#webhooks_message, etl_channel)

    return supehr_logger


def finalize():
    ''' Closes database connection and prints finish to log '''

    t_end_exe = datetime.utcnow()
    print("Total ETL execution time: " + str(t_end_exe - t_begin_exe))

    print("\nETL completed at " + strftime("%Y-%m-%d %H:%M:%S"))
    supehr_logger.info("\nETL_id:" + ETL_guid + " completed at " + strftime("%Y-%m-%d %H:%M:%S"))
    supehr_logger.info("\nTotal ETL execution time: " + str(t_end_exe - t_begin_exe))

    # Log to #webhooks
    #webhooks_message = "\nETL_id: " + ETL_guid + "\n *COMPLETED* at " + strftime("%Y-%m-%d %H:%M:%S")
    #webhooks_message += "\nTotal ETL execution time: " + str(t_end_exe - t_begin_exe)
    #webhooks(#webhooks_message,etl_channel)

   # if executing_statement not in executing_blacklist:
        # Log to #webhooks
        #webhooks('Paver finished: ' + executing_statement + ' on ' + constants.server_addr + '\n' + "ETL ID:" + ETL_guid, etl_channel)

# Track the total time of ETL. This will be useful to print out
# at the end 
t_begin_exe = datetime.utcnow()

# Database connection parameters    
username = constants.db_username
pw = constants.db_password
db = constants.db_name
server_address = constants.server_addr

# Global Variables
ETL_guid = uuid.uuid4().hex
executing_statement = sys.argv[1]
executing_blacklist = ['help', 'status']

# You can change this to be any number of default channels
etl_channel = '<default channel here>'

if executing_statement not in executing_blacklist:
    # Make sure the logging directory exists
    make_sure_log_directory_exists()
    # Start a logger
    supehr_logger = init_logging()
    # Log to #webhooks
    #webhooks('Paver started: ' + executing_statement + ' on ' + constants.server_addr + '\n' + "ETL ID:" + ETL_guid, etl_channel)
