#!/usr/bin/python

'''
Python does not add the current working directory to sys.path. This function
adds all packages to the path variable so imports can be resolved.
'''


def _paver_import_workaround():
    import os, sys
    sys.path.append(os.path.dirname(__file__))


_paver_import_workaround()

import os
import subprocess

from paver.easy import task, needs, no_help

from db_utils import finalize

'''
pavement.py is the engine of the SUPER ETL pipeline for i2b2 at Weill Cornell Medicine, powered by
paver. paver is a Python module that schedules and tracks dependencies through tasks. A paver task 
is a Python method with the @task annotation above it. This annotation, also known as a decorator, 
enables other operations within the paver module to be acted upon it. paver tasks can have dependencies,
which are other paver tasks that must be called prior to execution of the current task.
This is indicated with the @needs annotation.

Tasks are logically separated into Python files by i2b2 project, and within these scripts 
there are methods to build the ontologies and the data/views of a project.
Paver handles the complex dependencies created by the ETL pipeline and runs the SQL scripts in 
a specified order to produce a result. The result is often described by the name of the method, 
which reflects what project it runs on and what it does.

paver tasks can do anything from preprocess data to running reports on completed data marts to
creating indexes on source data. It is intended to be a tool to automate all portions of the
i2b2 ETL.

All paver tasks can be listed from the command prompt by issuing 'paver help', except for methods annotated 
with @no_help.Hidden methods annotated with @no_help should not be used unless you know what you are doing with them.
 All paver tasks in the pavement.py file can be executed with the command 'paver method_name'. All paver tasks
within individual project python files can be executed with the command 'paver <project>.<method_name>'.

'''
@task
@needs(['test.test_statement'])
def test_out_script():
    ''' Runs a test version of a SQL Script '''
    pass


@task
def status():
    ''' Prints the status of any running paver jobs '''
    process = subprocess.Popen("ps -ef | grep paver", shell=True, stdout=subprocess.PIPE)
    stdout_list = str(process.communicate()[0]).split('\\n')
    print('USER\tPID\tPPID\tCPU\tP\tTTY\tCPUT\tTASK')

    for x in range(0, len(stdout_list)):

        '''Format the last element and then insert a new line'''
        if (x % 8 == 0 and x != 0):
            if stdout_list[x].find("'/usr/bin/python /usr/bin/paver'") == -1:
                print(stdout_list[x] + '\n')
        else:
            if stdout_list[x].find("'/usr/bin/python /usr/bin/paver'") == -1:
                print(stdout_list[x] + '\t')
